from __future__ import annotations
from enum import Enum
import logging
import re
from typing import List

from openpectus.lang.model.pprogram import (
    PCommandWithDuration,
    PComment,
    PErrorInstruction,
    PNode,
    PProgram,
    PBlank,
    PBlock,
    PEndBlock,
    PEndBlocks,
    PWatch,
    PAlarm,
    PCommand,
    PMark,
)

from openpectus.lang.exec.tags import TagCollection
from openpectus.lang.exec.units import are_comparable
from openpectus.lang.exec.commands import CommandCollection
from openpectus.lang.exec.pinterpreter import PNodeVisitor
from openpectus.lang import float_re, unit_re


logging.basicConfig(format=' %(name)s :: %(levelname)-8s :: %(message)s')
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class AnalyzerItemType(Enum):
    HINT = 'HINT',
    INFO = 'INFO',
    WARNING = 'WARNING',
    ERROR = 'ERROR'


class AnalyzerItem():
    def __init__(self,
                 id: str,
                 message: str,
                 node: PNode | None,
                 type: AnalyzerItemType,
                 description: str = "") -> None:

        self.id: str = id
        self.message: str = message
        self.description: str = description
        self.type: AnalyzerItemType = type
        self.node: PNode | None = node


class AnalyzerVisitorBase(PNodeVisitor):
    """ Override by specific feature analyzers. """

    def __init__(self) -> None:
        super().__init__()
        self.items: List[AnalyzerItem] = []

    def add_item(self, item: AnalyzerItem):
        self.items.append(item)

    def visit_children(self, children: List[PNode] | None):
        if children is not None:
            for child in children:
                self.visit(child)

    def visit_PProgram(self, node: PProgram):
        self.visit_children(node.children)

    def visit_PBlank(self, node: PBlank):
        pass

    def visit_PMark(self, node: PMark):
        pass

    def visit_PBlock(self, node: PBlock):
        self.visit_children(node.children)

    def visit_PEndBlock(self, node: PEndBlock):
        pass

    def visit_PEndBlocks(self, node: PEndBlocks):
        pass

    def visit_PWatch(self, node: PWatch):
        self.visit_children(node.children)

    def visit_PAlarm(self, node: PAlarm):
        self.visit_children(node.children)

    def visit_PCommand(self, node: PCommand):
        self.visit_children(node.children)

    def visit_PCommandWithDuration(self, node: PCommandWithDuration):
        self.visit_children(node.children)

    def visit_PErrorInstruction(self, node: PErrorInstruction):
        self.visit_children(node.children)

    def visit_PComment(self, node: PComment):
        pass

# --- Enriching analyzers ---

class ConditionEnrichAnalyzer(AnalyzerVisitorBase):
    """ Analyzer that enriches parsed conditions with tag name, value and unit """
    def __init__(self) -> None:
        super().__init__()

    def visit_PWatch(self, node: PWatch):
        self.enrich_condition(node)
        return super().visit_PWatch(node)

    def visit_PAlarm(self, node: PAlarm):
        self.enrich_condition(node)
        return super().visit_PAlarm(node)

    def enrich_condition(self, node: PAlarm | PWatch):
        if node.condition is None:
            return
        c = node.condition
        if c.op is None or c.op not in ['<', '<=', '>', '>=', '==', '=', '!=']:
            return
        if c.lhs is None or c.lhs.strip() == '' or c.rhs is None or c.rhs.strip() == '':
            return
        c.tag_name = c.lhs

        re_rhs = '^' + '(?P<float>' + float_re + ')' + '\\s*' + '(?P<unit>' + unit_re + ')' + '$'
        re_rhs_no_unit = '^' + '(?P<float>' + float_re + ')' + '\\s*$'
        match = re.search(re_rhs, c.rhs)
        if match:  # float with unit
            c.tag_value = match.group('float')
            c.tag_value_numeric = float(c.tag_value or "")
            c.tag_unit = match.group('unit')
            c.error = False
        else:
            match = re.search(re_rhs_no_unit, c.rhs)
            if match:  # float without unit
                c.tag_value = match.group('float')
                c.tag_value_numeric = float(c.tag_value or "")
                c.error = False
            else:  # str
                c.tag_value = c.rhs
                c.error = False


class DurationEnrichAnalyzer(AnalyzerVisitorBase):
    """ Analyzer that enriches parsed durations with time and unit """
    def __init__(self) -> None:
        super().__init__()

    def visit_PCommandWithDuration(self, node: PCommandWithDuration):
        self.enrich_duration(node)
        return super().visit_PCommandWithDuration(node)

    def enrich_duration(self, node: PCommandWithDuration):
        if node.duration is None:
            return
        d = node.duration

        re_time_unit = '^' + '(?P<float>' + float_re + ')' + '\\s*' + '(?P<unit>' + unit_re + ')' + '$'
        # re_time = '^' + '(?P<float>' + float_re + ')' + '\\s*$'
        match = re.search(re_time_unit, d.duration_str)
        if match:
            d.time = float(match.group('float'))
            d.unit = match.group('unit')
            d.error = False
        # Removed in #437 although there was a test specifically for this.
        # Cannot think of a situation where as duration would be valid without a unit.
        # else:
        #     match = re.search(re_time, d.duration_str)
        #     if match:
        #         d.time = float(match.group('float'))
        #         d.error = False


class EnrichAnalyzer():
    """ Facade that combines the enrich analyzers into a single analyzer. """

    def __init__(self) -> None:
        super().__init__()
        self.analyzers = [
            ConditionEnrichAnalyzer(),
            DurationEnrichAnalyzer(),
        ]

    def analyze(self, program: PProgram):
        for analyzer in self.analyzers:
            analyzer.visit(program)


# --- Check analyzers ---

class UnreachableCodeCheckAnalyzer(AnalyzerVisitorBase):
    """ Analyser that checks for unreachable code """

    def create_item(self, node: PNode):
        return AnalyzerItem(
            "UnreachableCode",
            "Unreachable code",
            node,
            AnalyzerItemType.WARNING,
            "There is no path to this code."
        )

    def visit_PBlock(self, node: PBlock):
        has_end = False
        if node.children is not None:
            for child in node.children:
                if isinstance(child, (PEndBlock, PEndBlocks)):
                    has_end = True
                    continue
                if has_end:
                    self.add_item(self.create_item(child))


class InfiniteBlockCheckAnalyzer(AnalyzerVisitorBase):
    """ Analyser that checks for non-terminated blocks """

    def create_item(self, node: PNode):
        return AnalyzerItem(
            "InfiniteBlock",
            "Infinite block",
            node,
            AnalyzerItemType.WARNING,
            "This block will run indefinitely, as there are no End block or End blocks to terminate it."
        )

    def __init__(self) -> None:
        super().__init__()
        self.has_global_end = False
        self.requires_global_end: List[PNode] = []

    def check_global_end_block(self, node: PEndBlock | PEndBlocks):
        p = node.parent
        while p is not None:
            if isinstance(p, (PWatch, PAlarm)):
                if isinstance(p.parent, PProgram):
                    self.has_global_end = True
                    break
            p = p.parent

    def check_local_end_block(self, node: PBlock):
        for child in node.get_child_nodes(recursive=True):
            if isinstance(child, (PEndBlock, PEndBlocks)):
                return True
        return False

    def visit_PProgram(self, node: PProgram):
        super().visit_PProgram(node)

        if len(self.requires_global_end) > 0 and not self.has_global_end:
            for n in self.requires_global_end:
                self.add_item(self.create_item(n))

    def visit_PBlock(self, node: PBlock):
        super().visit_PBlock(node)

        has_end = self.check_local_end_block(node)
        # has_end = False
        # if node.children is not None:
        #     for child in node.children:
        #         if isinstance(child, (PEndBlock, PEndBlocks)):
        #             has_end = True
        #             break
        if not has_end:
            self.requires_global_end.append(node)

    def visit_PEndBlock(self, node: PEndBlock):
        self.check_global_end_block(node)

    def visit_PEndBlocks(self, node: PEndBlocks):
        self.check_global_end_block(node)


class ConditionCheckAnalyzer(AnalyzerVisitorBase):
    """ Analyser that checks whether conditions are present and refer to valid tag names """

    def __init__(self, tags: TagCollection) -> None:
        super().__init__()
        self.tags = tags

    def visit_PWatch(self, node: PWatch):
        self.analyze_condition(node)
        super().visit_PWatch(node)

    def visit_PAlarm(self, node: PAlarm):
        self.analyze_condition(node)
        super().visit_PAlarm(node)

    def analyze_condition(self, node: PWatch | PAlarm):
        if node.condition is None:
            self.add_item(AnalyzerItem(
                "ConditionMissing",
                "Condition missing",
                node,
                AnalyzerItemType.ERROR,
                "A condition is required"
            ))
            return
        condition = node.condition

        tag_name = condition.tag_name
        if tag_name is None or tag_name.strip() == '':
            self.add_item(AnalyzerItem(
                "MissingTag",
                "Missing tag",
                node,
                AnalyzerItemType.ERROR,
                "A condition must start with a tag name"
            ))
            return

        if not self.tags.has(tag_name):
            self.add_item(AnalyzerItem(
                "UndefinedTag",
                "Undefined tag",
                node,
                AnalyzerItemType.ERROR,
                f"The tag name '{tag_name}' is not valid"
            ))
            return

        tag = self.tags.get(tag_name)

        if tag.unit is None and condition.tag_unit is not None:
            self.add_item(AnalyzerItem(
                "UnexpectedUnit",
                "Unexpected tag unit",
                node,
                AnalyzerItemType.ERROR,
                f"Unit '{condition.tag_unit}' was specified for a tag with no unit"
            ))
            return

        if tag.unit is not None and condition.tag_unit is None:
            self.add_item(AnalyzerItem(
                "MissingUnit",
                "Missing tag unit",
                node,
                AnalyzerItemType.ERROR,
                "The tag requires that a unit is provided"
            ))
            return

        if tag.unit is not None:
            assert condition.tag_unit is not None
            if not are_comparable(tag.unit, condition.tag_unit):
                self.add_item(AnalyzerItem(
                    "IncompatibleUnits",
                    "Incompatible units",
                    node,
                    AnalyzerItemType.ERROR,
                    f"The tag unit '{tag.unit}' is not compatible with the provided unit '{condition.tag_unit}'"
                ))
                return


class CommandCheckAnalyzer(AnalyzerVisitorBase):
    """ Analyzer that checks commands and arguments. """

    def __init__(self, commands: CommandCollection) -> None:
        super().__init__()
        self.commands = commands

    def visit_PCommand(self, node: PCommand):
        assert node.name is not None
        name = node.name

        if not self.commands.has(name):
            self.add_item(AnalyzerItem(
                "UndefinedCommand",
                "Undefined command",
                node,
                AnalyzerItemType.ERROR,
                f"The command name '{name}' is not valid"
            ))
            return

        command = self.commands.get(name)
        if not command.validate_args(node.args):
            self.add_item(AnalyzerItem(
                "CommandArgsInvalid",
                "Invalid command arguments",
                node,
                AnalyzerItemType.ERROR,
                f"The command argument '{node.args}' is not valid"
            ))
            return


class SemanticCheckAnalyzer():
    """ Facade that combines the check analyzers into a single analyzer. """

    def __init__(self, tags: TagCollection, commands: CommandCollection) -> None:
        super().__init__()
        self.items: List[AnalyzerItem] = []
        self.analyzers = [
            UnreachableCodeCheckAnalyzer(),
            InfiniteBlockCheckAnalyzer(),
            ConditionCheckAnalyzer(tags),
            CommandCheckAnalyzer(commands)
        ]

    def analyze(self, program: PProgram):
        for analyzer in self.analyzers:
            analyzer.visit(program)
            self.items.extend(analyzer.items)

    @property
    def errors(self) -> List[AnalyzerItem]:
        return list([i for i in self.items if i.type == AnalyzerItemType.ERROR])

    @property
    def warnings(self) -> List[AnalyzerItem]:
        return list([i for i in self.items if i.type == AnalyzerItemType.ERROR])



import os
import unittest
from datetime import datetime

import openpectus.aggregator.data.models as DMdl
import openpectus.aggregator.models as Mdl
import openpectus.aggregator.routers.dto as Dto
from alembic import command
from alembic.config import Config
from openpectus.aggregator.data import database
from openpectus.aggregator.data.repository import RecentRunRepository
from openpectus.aggregator.routers.serialization import deserialize, serialize
from sqlalchemy import select
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import Mapped, mapped_column


def configure_db():
    database.configure_db("sqlite:///:memory:")


def create_db():
    # we can't use the migrations scripts, as in the migrate_db function, but must create the tables here on the fly,
    # due to the TestModel only being registered here, and thus not being migrated by the migration scripts
    assert database._engine is not None
    DMdl.DBModel.metadata.create_all(database._engine)


def migrate_db(alembic_cfg: Config):
    command.upgrade(alembic_cfg, 'head')


def get_modified_alembic_config():
    aggregator_prefix = os.path.join(os.path.dirname(__file__), "../../aggregator")
    alembic_cfg = Config(os.path.join(aggregator_prefix, "alembic.ini"))
    existing_script_location = alembic_cfg.get_main_option('script_location')
    assert existing_script_location is not None
    prefixed_script_location = os.path.join(aggregator_prefix, existing_script_location)
    alembic_cfg.set_main_option('script_location', prefixed_script_location)
    return alembic_cfg


def init_db():
    configure_db()
    create_db()


class DatabaseTest(unittest.TestCase):
    class TestModel(DMdl.DBModel):
        __tablename__ = "TestModel"

        name: Mapped[str] = mapped_column()

    def test_create_db(self):
        configure_db()

        stmt = select(DatabaseTest.TestModel)

        # select() raises when db does not exist
        with self.assertRaises(OperationalError):
            with database.create_scope():
                session = database.scoped_session()
                result = session.execute(stmt)
                self.assertIsNotNone(result)

        # create the tables
        create_db()

        # now execute works
        with database.create_scope():
            session = database.scoped_session()
            result = session.execute(stmt)
            self.assertIsNotNone(result)


    def test_no_missing_migration(self):
        db_filename = os.path.join(os.path.dirname(__file__), 'temp_db')
        configure_db()
        alembic_cfg = get_modified_alembic_config()
        alembic_cfg.set_main_option('sqlalchemy.url', "sqlite:///"+db_filename)

        migrate_db(alembic_cfg)
        # command.current(alembic_cfg, True) # useful for debugging
        command.check(alembic_cfg)
        os.remove(db_filename)

    def test_create_row_scoped(self):
        init_db()

        model = DatabaseTest.TestModel()
        model.name = "foo"

        self.assertEqual(model.id, None)

        # cannot access scope when there is no scope (contextmanager)
        with self.assertRaises(Exception):
            with database.scoped_session():
                pass

        with database.create_scope():
            session = database.scoped_session()
            session.add(model)
            session.commit()
            # refresh is required ...
            session.refresh(model)

        # ... to access updated property outside of the session
        self.assertEqual(model.id, 1)

        stmt = select(DatabaseTest.TestModel).where(DatabaseTest.TestModel.id == 1)
        with database.create_scope():
            session = database.scoped_session()
            result = session.execute(stmt)
            self.assertEqual(1, len(result.all()))

    def test_scope_session_identities(self):
        init_db()

        # same scope provides same session
        with database.create_scope():
            sess1 = database.scoped_session()
            id1 = id(sess1)
            sess2 = database.scoped_session()
            id2 = id(sess2)
            self.assertIs(sess1, sess2)
        self.assertEqual(id1, id2)

        # different scope provides different session
        with database.create_scope():
            sess3 = database.scoped_session()
            id3 = id(sess3)

        self.assertNotEqual(id1, id3)


class SerializationTest(unittest.TestCase):
    """ Test custom serialization for dto's used when serializing dto's as json database fields. """

    def test_json_serialization(self):
        data = Dto.ControlState(is_running=True, is_holding=False, is_paused=True)
        json_dict = serialize(data)
        self.assertEqual({"_type": "ControlState", 'is_running': True, 'is_holding': False, "is_paused": True}, json_dict)

    def test_json_deserialization(self):
        data = Dto.ControlState(is_running=True, is_holding=False, is_paused=True)
        json_dict = serialize(data)
        instance = deserialize(json_dict)
        self.assertIsNotNone(instance)
        self.assertIsInstance(instance, Dto.ControlState)
        self.assertEqual(instance.is_running, True)  # type: ignore

    # Enums not working. Not sure yet if they need to
    # def test_json_serialization_enum(self):
    #     data = D.UserRole.ADMIN
    #     json_dict = serialize(data)
    #     self.assertEqual({"_type": "UserRole"}, json_dict)

    def test_json_serialization_complex(self):
        complex_data = Dto.ProcessUnitState.InProgress(state=Dto.ProcessUnitStateEnum.IN_PROGRESS, progress_pct=87)
        json_str = serialize(complex_data)
        self.assertEqual({"_type": "ProcessUnitState.InProgress", "state": "in_progress", "progress_pct": 87}, json_str)

    def test_json_deserialization_complex(self):
        complex_data = Dto.ProcessUnitState.InProgress(state=Dto.ProcessUnitStateEnum.IN_PROGRESS, progress_pct=87)
        json_dict = serialize(complex_data)
        instance = deserialize(json_dict)
        self.assertIsNotNone(instance)
        self.assertIsInstance(instance, Dto.ProcessUnitState.InProgress)
        self.assertEqual(instance.progress_pct, 87)  # type: ignore


class RepositoryTest(unittest.TestCase):
    def test_can_insert(self):
        init_db()

        # at the data level we can insert any kind of junk
        entity = DMdl.RecentRun()
        entity.engine_id = "my_eng_id"
        entity.engine_computer_name = "my_computer_name"
        entity.uod_name = "my_uod_name"
        entity.uod_filename = "my_uod_filename.py"
        entity.uod_author_name = "my_uod_author_name"
        entity.uod_author_email = ""
        entity.run_id = 'a run id'
        entity.started_date = datetime.now()
        entity.completed_date = datetime.now()
        entity.engine_version = '0.0.1'
        entity.engine_hardware_str = 'hardware'
        entity.aggregator_version = '0.0.1'
        entity.aggregator_computer_name = 'aggregator computer name'

        entity_id = 0

        with database.create_scope():
            session = database.scoped_session()
            session.add(entity)
            session.commit()
            entity_id = entity.id

        self.assertEqual(entity_id, 1)

        with database.create_scope():
            session = database.scoped_session()
            repo = RecentRunRepository(session)

            created_entity = repo.get_by_run_id(entity.run_id)
            assert created_entity is not None
            self.assertEqual("my_computer_name", created_entity.engine_computer_name)
            self.assertEqual("my_uod_name", created_entity.uod_name)

    def test_can_update(self):
        init_db()

        entity = DMdl.RecentRun()
        entity.engine_id = "my_eng_id"
        entity.engine_computer_name = "my_computer_name"
        entity.uod_name = "my_uod_name"
        entity.uod_filename = "my_uod_filename.py"
        entity.uod_author_name = "my_uod_author_name"
        entity.uod_author_email = "my_uod_email"
        entity.run_id = 'a run id'
        entity.started_date = datetime.now()
        entity.completed_date = datetime.now()
        entity.engine_version = '0.0.1'
        entity.engine_hardware_str = 'hardware'
        entity.aggregator_version = '0.0.1'
        entity.aggregator_computer_name = 'aggregator computer name'

        entity_id = 0

        with database.create_scope():
            session = database.scoped_session()
            session.add(entity)
            session.commit()
            entity_id = entity.id

        self.assertEqual(1, entity_id)

        with database.create_scope():
            session = database.scoped_session()
            repo = RecentRunRepository(session)

            created_entity = repo.get_by_run_id(entity.run_id)
            assert created_entity is not None
            created_entity.engine_computer_name = "updated_computer_name"
            session.commit()

        with database.create_scope():
            session = database.scoped_session()
            repo = RecentRunRepository(session)

            updated_entity = repo.get_by_run_id(entity.run_id)
            assert updated_entity is not None
            self.assertEqual("updated_computer_name", updated_entity.engine_computer_name)
            self.assertEqual("my_uod_name", updated_entity.uod_name)

    def test_can_find(self):
        init_db()

        entity = DMdl.RecentRun()
        entity.engine_id = "my_eng_id"
        entity.run_id = 'a run id'
        entity.engine_computer_name = "my_computer_name"
        entity.uod_name = "my_uod_name"
        entity.uod_filename = "my_uod_filename.py"
        entity.uod_author_name = "my_uod_author_name"
        entity.uod_author_email = "my_uod_email"
        entity.started_date = datetime.now()
        entity.completed_date = datetime.now()
        entity.contributors = ['foo', 'bar']
        entity.engine_version = '0.0.1'
        entity.engine_hardware_str = 'hardware'
        entity.aggregator_version = '0.0.1'
        entity.aggregator_computer_name = 'aggregator computer name'

        with database.create_scope():
            session = database.scoped_session()
            session.add(entity)
            session.commit()

        with database.create_scope():
            session = database.scoped_session()
            repo = RecentRunRepository(session)
            created_entity = repo.get_by_run_id("a run id")
            self.assertIsNotNone(created_entity)
            self.assertIsInstance(created_entity, DMdl.RecentRun)

    def test_json(self):
        init_db()

        entity = DMdl.RecentRun()
        entity.engine_id = "my_eng_id"
        entity.engine_computer_name = "my_computer_name"
        entity.uod_name = "my_uod_name"
        entity.uod_filename = "my_uod_filename.py"
        entity.uod_author_name = "my_uod_author_name"
        entity.uod_author_email = "my_uod_email"
        entity.contributors = ['foo', 'bar']
        entity.run_id = 'a run id'
        entity.started_date = datetime.now()
        entity.completed_date = datetime.now()
        entity.engine_version = '0.0.1'
        entity.engine_hardware_str = 'hardware'
        entity.aggregator_version = '0.0.1'
        entity.aggregator_computer_name = 'aggregator computer name'

        entity_id = 0
        with database.create_scope():
            session = database.scoped_session()
            session.add(entity)
            session.commit()
            entity_id = entity.id

        self.assertEqual(entity_id, 1)

        with database.create_scope():
            session = database.scoped_session()
            repo = RecentRunRepository(session)

            created_entity = repo.get_by_run_id(entity.run_id)
            assert created_entity is not None
            self.assertEqual(set(['foo', 'bar']), set(created_entity.contributors))

            # Must reassign with new instance for change to be persisted
            contributors = set(created_entity.contributors)
            contributors.add('baz')
            created_entity.contributors = list(contributors)
            session.commit()

        with database.create_scope():
            session = database.scoped_session()
            repo = RecentRunRepository(session)
            updated_entity = repo.get_by_run_id(entity.run_id)
            assert updated_entity is not None
            self.assertEqual(set(['foo', 'bar', 'baz']), set(updated_entity.contributors))

    def test_nested_json(self):
        init_db()

        run_id = 'fdsa'

        plot_configuration = Mdl.PlotConfiguration.empty()
        plot_configuration.process_value_names_to_annotate = ['x']

        recent_run_plot_configuration = DMdl.RecentRunPlotConfiguration()
        recent_run_plot_configuration.run_id = run_id
        recent_run_plot_configuration.plot_configuration = plot_configuration

        with database.create_scope():
            session = database.scoped_session()
            session.add(recent_run_plot_configuration)
            session.commit()

            repo = RecentRunRepository(session)
            from_db = repo.get_plot_configuration_by_run_id(run_id)
            dto: Dto.PlotConfiguration = Dto.PlotConfiguration.model_validate(from_db)
            self.assertIsInstance(dto, Dto.PlotConfiguration)
            self.assertEqual(dto.process_value_names_to_annotate, plot_configuration.process_value_names_to_annotate)


if __name__ == "__main__":
    unittest.main()



"""Add RecentEngine

Revision ID: 05c4bf88d0bb
Revises: d517729a9be0
Create Date: 2024-06-24 13:53:40.744147

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '05c4bf88d0bb'
down_revision: Union[str, None] = 'd517729a9be0'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('RecentEngines',
    sa.Column('engine_id', sa.String(), nullable=False),
    sa.Column('run_id', sa.String(), nullable=True),
    sa.Column('run_started', sa.DateTime(), nullable=True),
    sa.Column('run_stopped', sa.DateTime(), nullable=True),
    sa.Column('name', sa.String(), nullable=False),
    sa.Column('system_state', sa.String(), nullable=False),
    sa.Column('location', sa.String(), nullable=False),
    sa.Column('last_update', sa.DateTime(), nullable=False),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_RecentEngines')),
    sa.UniqueConstraint('engine_id', name=op.f('uq_RecentEngines_engine_id'))
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('RecentEngines')
    # ### end Alembic commands ###



from __future__ import annotations
import logging
import time
from typing import Any

from openpectus.engine.internal_commands import InternalCommandsRegistry, InternalEngineCommand
from openpectus.engine.models import EngineCommandEnum, MethodStatusEnum, SystemStateEnum
from openpectus.lang.exec.tags import SystemTagName
from openpectus.engine.engine import Engine


logger = logging.getLogger(__name__)


CANCEL_TIMEOUT_TICKS = 10

# Note: Classes in this module are auto-registered as internal engine commands by
# InternalCommandsRegistry during engine initialization.

def get_duration_end(tick_time: float, time: float, unit: str) -> float:
    if unit not in ['s', 'min', 'h']:
        raise ValueError(f"Wait argument unit must be a time unit, not '{unit}'")

    seconds = time
    if unit == 'min':
        seconds = 60 * time
    elif unit == 'h':
        seconds = 60 * 60 * time
    return tick_time + seconds


class StartEngineCommand(InternalEngineCommand):
    def __init__(self, engine: Engine, registry: InternalCommandsRegistry) -> None:
        super().__init__(EngineCommandEnum.START, registry)
        self.engine = engine

    def _run(self):
        e = self.engine
        if e._runstate_started:
            logger.warning("Cannot start when already running")
            self.fail()
        else:
            e._runstate_started = True
            e._runstate_started_time = time.time()
            e._runstate_paused = False
            e._runstate_holding = False
            run_id = e.set_run_id()
            e._system_tags[SystemTagName.SYSTEM_STATE].set_value(SystemStateEnum.Running, e._tick_time)
            e._system_tags[SystemTagName.METHOD_STATUS].set_value(MethodStatusEnum.OK, e._tick_time)
            e._system_tags[SystemTagName.RUN_TIME].set_value(0.0, e._tick_time)
            e._system_tags[SystemTagName.PROCESS_TIME].set_value(0.0, e._tick_time)
            e._system_tags[SystemTagName.RUN_COUNTER].set_value(0, e._tick_time)

            e._system_tags[SystemTagName.BLOCK_TIME].set_value(0.0, e._tick_time)
            e.block_times.clear()  # kinda hackish, tag should be self-contained

            e.emitter.emit_on_start(run_id)


class PauseEngineCommand(InternalEngineCommand):
    """ Pause execution of commands and time. Put output tags into safe state.

    See also Hold and Wait.
    """
    def __init__(self, engine: Engine, registry: InternalCommandsRegistry) -> None:
        super().__init__(EngineCommandEnum.PAUSE, registry)
        self.engine = engine
        self.duration_end_time : float | None = None

    def init_args(self, kvargs: dict[str, Any]):
        if "time" in kvargs.keys() and "unit" in kvargs.keys():
            time = float(kvargs.get("time", None))
            if time is None:
                raise ValueError("Invalid Pause arguments. Time is not valid")
            unit = kvargs.get("unit", None)
            if unit is None:
                raise ValueError("Invalid Pause arguments. Unit is not valid")
            self.duration_end_time = get_duration_end(self.engine._tick_time, time, unit)
        elif "time" in kvargs.keys() or "unit" in kvargs.keys():
            raise ValueError("Invalid Pause arguments. Specify either no duration arguments or both time and unit")

    def _run(self):
        e = self.engine
        e._runstate_paused = True
        e._system_tags[SystemTagName.SYSTEM_STATE].set_value(SystemStateEnum.Paused, e._tick_time)
        e._prev_state = e._apply_safe_state()

        if self.duration_end_time is not None:
            logger.debug("Pause duration set. Waiting to unpause.")
            while self.engine._tick_time < self.duration_end_time:
                yield
            logger.debug("Resuming using Unpause")
            UnpauseEngineCommand(self.engine, self._registry)._run()


class UnpauseEngineCommand(InternalEngineCommand):
    def __init__(self, engine: Engine, registry: InternalCommandsRegistry) -> None:
        super().__init__(EngineCommandEnum.UNPAUSE, registry)
        self.engine = engine

    def _run(self):
        e = self.engine

        e._runstate_paused = False
        if e._runstate_holding:
            e._system_tags[SystemTagName.SYSTEM_STATE].set_value(SystemStateEnum.Holding, e._tick_time)
        else:
            e._system_tags[SystemTagName.SYSTEM_STATE].set_value(SystemStateEnum.Running, e._tick_time)

        # pre-pause values are always applied on unpause, regardless of hold state.
        if e._prev_state is not None:
            e._apply_state(e._prev_state)
            e._prev_state = None
        else:
            logger.error("Failed to apply state prior to safe state. Prior state was not available")

        # TODO Consider how a corrected error should be handled. It depends on the cause
        # - Run time value is broken: Maybe trying again will fix it
        # - Code is broken and the user fixes it: The interpreter should resume from the error node
        # which can be complex and requires the full editing feature. For now, we just take the error
        # flag down and hope for the best
        e._system_tags[SystemTagName.METHOD_STATUS].set_value(MethodStatusEnum.OK, e._tick_time)


class HoldEngineCommand(InternalEngineCommand):
    """ Hold execution of commands and time, keeping ouput tags in their current state.

    See also Pause and Wait.
    """
    def __init__(self, engine: Engine, registry: InternalCommandsRegistry) -> None:
        super().__init__(EngineCommandEnum.HOLD, registry)
        self.engine = engine
        self.duration_end_time : float | None = None

    def init_args(self, kvargs: dict[str, Any]):
        if "time" in kvargs.keys() and "unit" in kvargs.keys():
            time = float(kvargs.get("time", None))
            if time is None:
                raise ValueError("Invalid Hold arguments. Time is not valid")
            unit = kvargs.get("unit", None)
            if unit is None:
                raise ValueError("Invalid Hold arguments. Unit is not valid")
            self.duration_end_time = get_duration_end(self.engine._tick_time, time, unit)
        elif "time" in kvargs.keys() or "unit" in kvargs.keys():
            raise ValueError("Invalid Hold arguments. Specify either no duration arguments or both time and unit")

    def _run(self):
        e = self.engine
        e._runstate_holding = True
        if not e._runstate_paused:  # Pause takes precedence
            e._system_tags[SystemTagName.SYSTEM_STATE].set_value(SystemStateEnum.Holding, e._tick_time)

        if self.duration_end_time is not None:
            logger.debug("Hold duration set. Waiting to unhold.")
            while self.engine._tick_time < self.duration_end_time:
                yield
            logger.debug("Resuming using Unhold")
            UnholdEngineCommand(self.engine, self._registry)._run()


class UnholdEngineCommand(InternalEngineCommand):
    def __init__(self, engine: Engine, registry: InternalCommandsRegistry) -> None:
        super().__init__(EngineCommandEnum.UNHOLD, registry)
        self.engine = engine

    def _run(self):
        e = self.engine
        e._runstate_holding = False
        if not e._runstate_paused:  # Pause takes precedence
            e._system_tags[SystemTagName.SYSTEM_STATE].set_value(SystemStateEnum.Running, e._tick_time)


class StopEngineCommand(InternalEngineCommand):
    def __init__(self, engine: Engine, registry: InternalCommandsRegistry) -> None:
        super().__init__(EngineCommandEnum.STOP, registry)
        self.engine = engine

    def _run(self):
        e = self.engine
        sys_state = e._system_tags[SystemTagName.SYSTEM_STATE]
        if sys_state.get_value() == SystemStateEnum.Stopped:
            logger.warning("Cannot stop when system state is already stopped")
            self.fail()
        elif sys_state.get_value() == SystemStateEnum.Restarting:
            logger.warning("Cannot stop when system state is restarting")
            self.fail()
        else:
            e._runstate_stopping = True
            e._cancel_uod_commands()
            yield
            timeout_at_tick = e._tick_number + CANCEL_TIMEOUT_TICKS
            while e.uod.has_any_command_instances():
                if e._tick_number > timeout_at_tick:
                    logger.warning("Timeout waiting for uod commands to cancel")
                    break
                yield
            e._finalize_uod_commands()

            logger.debug("All uod commands have completed execution. Stop will now complete.")
            e._runstate_started = False
            e._runstate_paused = False
            e._runstate_holding = False
            e._runstate_stopping = False
            e._system_tags[SystemTagName.METHOD_STATUS].set_value(MethodStatusEnum.OK, e._tick_time)

            e.emitter.emit_on_stop()

            e._system_tags[SystemTagName.SYSTEM_STATE].set_value(SystemStateEnum.Stopped, e._tick_time)
            e.clear_run_id()
            e._stop_interpreter()


class WaitEngineCommand(InternalEngineCommand):
    """ Pause execution of commands for the specified duration, keeping time running and output tags in their current state.

    See also Pause and Hold.
    """
    def __init__(self, engine: Engine, registry: InternalCommandsRegistry) -> None:
        super().__init__(EngineCommandEnum.WAIT, registry)
        self.engine = engine
        self.forced = False

    def init_args(self, kvargs: dict[str, Any]):
        if "time" in kvargs.keys() and "unit" in kvargs.keys():
            time = float(kvargs.get("time", None))
            if time is None:
                raise ValueError("Invalid Wait arguments. Time is not valid")
            unit = kvargs.get("unit", None)
            if unit is None:
                raise ValueError("Invalid Wait arguments. Unit is not valid")
            self.duration_end_time = get_duration_end(self.engine._tick_time, time, unit)
        else:
            raise ValueError("Invalid Wait arguments. A duration is required")

    def _run(self):
        self.engine._runstate_waiting = True
        start = self.engine._tick_time
        duration = self.duration_end_time - start
        while self.engine._tick_time < self.duration_end_time and not self.forced:
            if duration > 0:
                progress = (self.engine._tick_time - start) / duration
                self.set_progress(progress)
            yield
        self.engine._runstate_waiting = False

    def force(self):
        self.forced = True

class RestartEngineCommand(InternalEngineCommand):
    def __init__(self, engine: Engine, registry: InternalCommandsRegistry) -> None:
        super().__init__(EngineCommandEnum.RESTART, registry)
        self.engine = engine

    def _run(self):
        e = self.engine
        sys_state = e._system_tags[SystemTagName.SYSTEM_STATE]
        if sys_state.get_value() == SystemStateEnum.Stopped:
            logger.warning("Cannot restart when system state is Stopped")
            self.fail()
        elif sys_state.get_value() == SystemStateEnum.Restarting:
            logger.warning("Cannot restart when system state is Restarting")
            self.fail()
        else:
            logger.info("Restarting engine")
            sys_state.set_value(SystemStateEnum.Restarting, e._tick_time)

            e._runstate_stopping = True
            e._cancel_uod_commands()
            yield
            timeout_at_tick = e._tick_number + CANCEL_TIMEOUT_TICKS
            while e.uod.has_any_command_instances():
                if e._tick_number > timeout_at_tick:
                    logger.warning("Timeout waiting for uod commands to cancel")
                    break
                yield
            e._finalize_uod_commands()

            logger.debug("Restarting engine - uod commands have completed execution")
            e._runstate_started = False
            e._runstate_paused = False
            e._runstate_holding = False
            e._runstate_stopping = False
            e._system_tags[SystemTagName.SYSTEM_STATE].set_value(SystemStateEnum.Stopped, e._tick_time)

            e.emitter.emit_on_stop()

            e.clear_run_id()
            e._stop_interpreter()
            logger.info("Restarting engine - engine stopped")

            yield

            logger.info("Restarting engine - starting engine")
            # potentially a lot more engine state to reset
            e._runstate_started = True
            e._runstate_started_time = time.time()
            e._runstate_paused = False
            e._runstate_holding = False
            e._system_tags[SystemTagName.BLOCK_TIME].set_value(0.0, e._tick_time)
            e.block_times.clear()  # kinda hackish, tag should be self-contained

            run_id = e.set_run_id()
            e.emitter.emit_on_start(run_id)

            e._system_tags[SystemTagName.SYSTEM_STATE].set_value(SystemStateEnum.Running, e._tick_time)
            logger.info("Restarting engine complete")


class InfoEngineCommand(InternalEngineCommand):
    def __init__(self, engine: Engine, registry: InternalCommandsRegistry) -> None:
        super().__init__(EngineCommandEnum.INFO, registry)

    def _run(self):
        msg = self.kvargs.get("unparsed_args")
        logger.info(f"Info: {msg}")


class WarningEngineCommand(InternalEngineCommand):
    def __init__(self, engine: Engine, registry: InternalCommandsRegistry) -> None:
        super().__init__(EngineCommandEnum.WARNING, registry)

    def _run(self):
        msg = self.kvargs.get("unparsed_args")
        logger.warning(f"Warning: {msg}")


class ErrorEngineCommand(InternalEngineCommand):
    def __init__(self, engine: Engine, registry: InternalCommandsRegistry) -> None:
        super().__init__(EngineCommandEnum.ERROR, registry)

    def _run(self):
        msg = self.kvargs.get("unparsed_args")
        logger.error(f"Error: {msg}")



import re
import unittest

from openpectus.lang.exec.analyzer import (
    ConditionEnrichAnalyzer,
    DurationEnrichAnalyzer,
)
from openpectus.lang.model.pprogram import PCommandWithDuration, PCondition, PDuration, PWatch
from openpectus.lang import float_re, unit_re, identifier_re
from openpectus.test.engine.utility_methods import build_program as _build_program

def build_program_wo_analyzers(code: str):
    return _build_program(code, skip_enrich_analyzers=True)


class ConditionEnrichAnalyzerTest(unittest.TestCase):
    def test_watch_tag(self):
        program = build_program_wo_analyzers("""
Watch: counter > 0 mL
    Mark: b
""")
        analyzer = ConditionEnrichAnalyzer()
        watch = program.get_instructions()[0]
        assert isinstance(watch, PWatch)
        assert isinstance(watch.condition, PCondition)

        self.assertEqual(watch.condition.lhs, "counter")
        self.assertEqual(watch.condition.op, ">")
        self.assertEqual(watch.condition.rhs, "0 mL")

        self.assertEqual(watch.condition.tag_name, None)
        self.assertEqual(watch.condition.tag_value, None)
        self.assertEqual(watch.condition.tag_unit, None)

        analyzer.visit(program)

        self.assertEqual(watch.condition.tag_name, "counter")


    def test_watch_value_unit(self):
        program = build_program_wo_analyzers("""
Watch: counter > 0 mL
    Mark: b
""")
        analyzer = ConditionEnrichAnalyzer()
        watch = program.get_instructions()[0]
        assert isinstance(watch, PWatch) and isinstance(watch.condition, PCondition)
        analyzer.visit(program)

        self.assertEqual(watch.condition.tag_name, "counter")
        self.assertEqual(watch.condition.tag_value, "0")
        self.assertEqual(watch.condition.tag_unit, "mL")

    def test_float_re(self):
        def test(s):
            with self.subTest(s):
                m = re.match(float_re, s)
                self.assertIsNotNone(m)
        test("0")
        test("1.0")
        test("-0")
        test("-1")
        test("0.0")
        test("87,21")

    def test_unit_re(self):
        def test(s, expect_valid=True):
            with self.subTest(s):
                m = re.match('^' + unit_re + '$', s)
                if expect_valid:
                    self.assertIsNotNone(m)
                else:
                    self.assertIsNone(m)

        test("L")
        test("s")
        test("min")
        test("foo")
        test("%")
        test("kg/s")

        test("", False)
        test(".", False)
        test("0", False)
        test("0k", False)

    def test_identifier_re(self):
        def test(s, expect_valid=True):
            with self.subTest(s):
                m = re.match('^' + identifier_re + '$', s)
                if expect_valid:
                    self.assertIsNotNone(m)
                else:
                    self.assertIsNone(m)
        test("L")
        test("L2")
        test("L2")
        test("foo bar")
        test("foo2bar")
        test("foo bar baz")
        test("b 2")

        test("2", False)
        test("2b", False)
        test("2 b", False)
        test("", False)


    def test_watch_combinations(self):
        def test(code, expected_tag, expected_value, expected_unit):
            with self.subTest(code):
                program = build_program_wo_analyzers(code)
                analyzer = ConditionEnrichAnalyzer()
                watch = program.get_instructions()[0]
                assert isinstance(watch, PWatch) and isinstance(watch.condition, PCondition)
                analyzer.visit(program)

                self.assertEqual(watch.condition.tag_name, expected_tag)
                self.assertEqual(watch.condition.tag_value, expected_value)
                self.assertEqual(watch.condition.tag_unit, expected_unit)

        test("Watch: a > 0 s", "a", "0", "s")
        test("Watch: a >0", "a", "0", None)
        test("Watch: a > 0 ", "a", "0", None)
        test("Watch: foo bar > 0 mL ", "foo bar", "0", "mL")
        test("Watch: foo bar > 0   ", "foo bar", "0", None)


class DurationEnrichAnalyzerTest(unittest.TestCase):
    def test_pause_w_unit(self):
        program = build_program_wo_analyzers("Pause: 5 s")
        analyzer = DurationEnrichAnalyzer()
        pause = program.get_instructions()[0]
        assert isinstance(pause, PCommandWithDuration)
        assert isinstance(pause.duration, PDuration)

        self.assertEqual(pause.duration.duration_str, "5 s")
        self.assertEqual(pause.duration.time, None)
        self.assertEqual(pause.duration.unit, None)
        self.assertEqual(pause.duration.error, True)

        analyzer.visit(program)

        self.assertEqual(pause.duration.time, 5.0)
        self.assertEqual(pause.duration.unit, "s")
        self.assertEqual(pause.duration.error, False)

    def test_pause_wo_unit(self):
        program = build_program_wo_analyzers("Pause: 3.14")
        analyzer = DurationEnrichAnalyzer()
        pause = program.get_instructions()[0]
        assert isinstance(pause, PCommandWithDuration)
        assert isinstance(pause.duration, PDuration)

        self.assertEqual(pause.duration.duration_str, "3.14")
        self.assertEqual(pause.duration.time, None)
        self.assertEqual(pause.duration.unit, None)
        self.assertEqual(pause.duration.error, True)

        analyzer.visit(program)

        # Note: Changed in #437. Cannot see how duration could ever be valid without a unit
        # self.assertEqual(pause.duration.time, 3.14)
        # self.assertEqual(pause.duration.unit, None)
        # self.assertEqual(pause.duration.error, False)
        self.assertEqual(pause.duration.duration_str, "3.14")
        self.assertEqual(pause.duration.time, None)
        self.assertEqual(pause.duration.unit, None)
        self.assertEqual(pause.duration.error, True)



from typing import Awaitable, Callable, Any, TypeVar

import openpectus.protocol.aggregator_messages as AM
import openpectus.protocol.engine_messages as EM
import openpectus.protocol.messages as M

# --------- Dispatch interfaces ------------

AGGREGATOR_RPC_WS_PATH = "/engine-rpc"
AGGREGATOR_REST_PATH = "/engine-rest"
AGGREGATOR_HEALTH_PATH = "/health"
AGGREGATOR_AUTH_CONFIG_PATH = "/auth/config"
MessageHandler = Callable[[Any], Awaitable[M.MessageBase]]

AnyMessageType = TypeVar("AnyMessageType", bound=M.MessageBase)
EngineMessageType = TypeVar("EngineMessageType", bound=EM.EngineMessage)
AggregatorMessageType = TypeVar("AggregatorMessageType", bound=AM.AggregatorMessage)



from typing import List

import openpectus.aggregator.models as Mdl
import openpectus.aggregator.routers.dto as Dto
import openpectus.aggregator.data.models as Db
from fastapi import APIRouter, HTTPException
from fastapi.responses import StreamingResponse
from openpectus.aggregator.csv_generator import generate_csv_string
from openpectus.aggregator.data import database
from openpectus.aggregator.data.repository import PlotLogRepository, RecentRunRepository
from openpectus.aggregator.routers.auth import UserRolesValue, has_access
from starlette.status import HTTP_403_FORBIDDEN, HTTP_404_NOT_FOUND

router = APIRouter(tags=["recent_runs"], prefix='/recent_runs')


def get_recent_run_or_fail(user_roles: UserRolesValue, run_id: str) -> Db.RecentRun:
    repo = RecentRunRepository(database.scoped_session())
    recent_run = repo.get_by_run_id(run_id)
    if recent_run is None:
        raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail='Recent Run not found')
    if not has_access(recent_run, user_roles):
        raise HTTPException(status_code=HTTP_403_FORBIDDEN, detail={'missing_roles': recent_run.required_roles})
    return recent_run


@router.get("/", response_model_exclude_none=True)
def get_recent_runs(user_roles: UserRolesValue) -> List[Dto.RecentRun]:
    repo = RecentRunRepository(database.scoped_session())
    return list(map(Dto.RecentRun.model_validate, filter(lambda rr: has_access(rr, user_roles), repo.get_all())))


@router.get("/{run_id}", response_model_exclude_none=True)
def get_recent_run(user_roles: UserRolesValue, run_id: str) -> Dto.RecentRun:
    recent_run = get_recent_run_or_fail(user_roles, run_id)
    return Dto.RecentRun.model_validate(recent_run)


@router.get('/{run_id}/method-and-state', response_model_exclude_none=True)
def get_recent_run_method_and_state(user_roles: UserRolesValue, run_id: str) -> Dto.MethodAndState:
    get_recent_run_or_fail(user_roles, run_id)
    repo = RecentRunRepository(database.scoped_session())
    method_and_state = repo.get_method_and_state_by_run_id(run_id)
    if method_and_state is None:
        raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail='Method and state not found')
    return Dto.MethodAndState.model_validate(method_and_state)


@router.get('/{run_id}/run_log', response_model_exclude_none=True)
def get_recent_run_run_log(user_roles: UserRolesValue, run_id: str) -> Dto.RunLog:
    repo = RecentRunRepository(database.scoped_session())
    get_recent_run_or_fail(user_roles, run_id)
    run_log_db = repo.get_run_log_by_run_id(run_id)
    if run_log_db is None:
        raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail='Run Log not found')
    run_log_mdl = Mdl.RunLog.model_validate(run_log_db)
    return Dto.RunLog(lines=list(map(Dto.RunLogLine.from_model, run_log_mdl.lines)))


@router.get('/{run_id}/plot_configuration', response_model_exclude_none=True)
def get_recent_run_plot_configuration(user_roles: UserRolesValue, run_id: str) -> Dto.PlotConfiguration:
    repo = RecentRunRepository(database.scoped_session())
    get_recent_run_or_fail(user_roles, run_id)
    plot_configuration = repo.get_plot_configuration_by_run_id(run_id)
    if plot_configuration is None:
        raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail='Plot Configuration not found')
    return Dto.PlotConfiguration.model_validate(plot_configuration)


@router.get('/{run_id}/plot_log', response_model_exclude_none=True)
def get_recent_run_plot_log(user_roles: UserRolesValue, run_id: str) -> Dto.PlotLog:
    get_recent_run_or_fail(user_roles, run_id)
    plot_repo = PlotLogRepository(database.scoped_session())
    plot_log_model = plot_repo.get_plot_log(run_id)
    if plot_log_model is None:
        raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail='Plot Log not found')
    return Dto.PlotLog.model_validate(plot_log_model)


@router.get('/{run_id}/csv_json', response_model_exclude_none=True)
def get_recent_run_csv_json(user_roles: UserRolesValue, run_id: str) -> Dto.RecentRunCsv:
    recent_run = get_recent_run_or_fail(user_roles, run_id)

    plot_log_repo = PlotLogRepository(database.scoped_session())
    plot_log_model = plot_log_repo.get_plot_log(run_id)
    if plot_log_model is None:
        raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail='Plot Log not found')

    plot_log = Dto.PlotLog.model_validate(plot_log_model)
    csv_string = generate_csv_string(plot_log, Dto.RecentRun.model_validate(recent_run))
    return Dto.RecentRunCsv(filename=f'RecentRun-{run_id}.csv', csv_content=csv_string.getvalue())


@router.get('/{run_id}/error_log', response_model_exclude_none=True)
def get_recent_run_error_log(user_roles: UserRolesValue, run_id: str) -> Dto.AggregatedErrorLog:
    get_recent_run_or_fail(user_roles, run_id)
    repo = RecentRunRepository(database.scoped_session())
    error_log = repo.get_error_log_by_run_id(run_id)
    if error_log is None:
        raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail='Error Log not found')
    return Dto.AggregatedErrorLog.from_model(Mdl.AggregatedErrorLog.model_validate(error_log))


@router.get('/{run_id}/csv_file', response_class=StreamingResponse, response_model_exclude_none=True)
def get_recent_run_csv_file(user_roles: UserRolesValue, run_id: str) -> StreamingResponse:
    get_recent_run_or_fail(user_roles, run_id)
    file_content = 'some;CSV;here\nand;more;here'
    file_name = f'RecentRun-{run_id}.csv'
    return StreamingResponse(
        content=iter([file_content]),
        media_type='text/csv',
        headers={'Content-Disposition': f'attachment;filename="{file_name}"'}
    )



import unittest

from antlr4.tree.Tree import ErrorNode as antlrErrorNode

from openpectus.lang.grammar.codegen.pcodeParser import ParserRuleContext, pcodeParser
from openpectus.lang.grammar.pgrammar import PGrammar


def parse(s):
    p = PGrammar()
    p.parse(s)
    return p


def get_first_child(ctx: ParserRuleContext, type) -> ParserRuleContext | None:
    """ Return the first child of the specified type (recursively) if one exists. Else return None"""
    if isinstance(ctx, type):
        return ctx
    if hasattr(ctx, "getChildren"):
        for child in ctx.getChildren():
            if isinstance(child, type):
                return child
            val = get_first_child(child, type)
            if val:
                return val


class ParserTest(unittest.TestCase):

    def assertContextHasNoChildError(self, ctx: ParserRuleContext):
        # if hasattr(ctx, 'exception'):
        #     self.assertIsNone(ctx.exception)
        if hasattr(ctx, "getChildren"):
            for child in ctx.getChildren():
                self.assertNotIsInstance(
                    child,
                    antlrErrorNode,
                    f"Child of context '{ctx.getText()}' with type '{type(ctx).__name__}' was unexpectedly an ErrorNode")
                self.assertContextHasNoChildError(child)

    def test_mark(self):
        p = parse("Mark:   a  ")
        c = p.parser.mark()  # type: ignore
        self.assertIsNotNone(c)
        # p.printSyntaxTree(c)
        self.assertIsInstance(c, pcodeParser.MarkContext)
        self.assertContextHasNoChildError(c)

    def test_block(self):
        p = parse("Block: foo")
        c = p.parser.block()  # type: ignore
        self.assertIsNotNone(c)
        # p.printSyntaxTree(c)
        self.assertIsInstance(c, pcodeParser.BlockContext)
        self.assertContextHasNoChildError(c)

        p = parse("Block: foo")
        c = p.parser.program()  # type: ignore
        self.assertIsNotNone(c)
        self.assertIsInstance(c, pcodeParser.ProgramContext)
        self.assertContextHasNoChildError(c)

    def test_end_block(self):
        p = parse("End block")
        c = p.parser.end_block()  # type: ignore
        self.assertIsNotNone(c)
        # p.printSyntaxTree(c)
        self.assertIsInstance(c, pcodeParser.End_blockContext)
        self.assertContextHasNoChildError(c)

    def test_end_blocks(self):
        p = parse("End blocks")
        c = p.parser.end_blocks()  # type: ignore
        self.assertIsNotNone(c)
        # p.printSyntaxTree(c)
        self.assertIsInstance(c, pcodeParser.End_blocksContext)
        self.assertContextHasNoChildError(c)

    def test_watch(self):
        p = parse("Watch")
        c = p.parser.watch()  # type: ignore
        self.assertIsNotNone(c)
        # p.printSyntaxTree(c)
        self.assertIsInstance(c, pcodeParser.WatchContext)
        self.assertContextHasNoChildError(c)

    def test_watch_unit(self):
        p = parse("Watch: A > 2 mL")
        c = p.parser.watch()  # type: ignore
        self.assertIsNotNone(c)
        # p.printSyntaxTree(c)
        self.assertIsInstance(c, pcodeParser.WatchContext)
        self.assertConditionValue(c.condition(), "A", ">", "2 mL")
        self.assertContextHasNoChildError(c)

    def test_watch_unitless(self):
        p = parse("Watch: A > 2")
        c = p.parser.watch()  # type: ignore
        self.assertIsNotNone(c)
        # p.printSyntaxTree(c)
        self.assertIsInstance(c, pcodeParser.WatchContext)
        self.assertConditionValue(c.condition(), "A", ">", "2")
        self.assertContextHasNoChildError(c)

    def test_watch_with_mark(self):
        p = parse("""
Watch: A > 2 mL
    Mark: a
""")
        program = p.parser.program()  # type: ignore
        watch = get_first_child(program, pcodeParser.WatchContext)
        self.assertIsInstance(watch, pcodeParser.WatchContext)
        self.assertIsWatchWithCondition(watch, "A > 2 mL")  # type: ignore
        self.assertConditionValue(watch.condition(), "A", ">", "2 mL")  # type: ignore

        mark = get_first_child(program, pcodeParser.MarkContext)
        self.assertIsInstance(mark, pcodeParser.MarkContext)

    def test_watch_w_inst_error_colon(self):
        p = parse("""
Watch:
    Mark: A""")
        program = p.parser.program()  # type: ignore
        # p.printSyntaxTree(program)
        watch = get_first_child(program, pcodeParser.WatchContext)
        assert isinstance(watch, pcodeParser.WatchContext)
        mark = get_first_child(program, pcodeParser.MarkContext)
        self.assertIsInstance(mark, pcodeParser.MarkContext)

    def test_watch_w_inst_error_colon_comment(self):
        p = parse("""
Watch: # foo
    Mark: A""")
        program = p.parser.program()  # type: ignore
        # p.printSyntaxTree(program)
        watch = get_first_child(program, pcodeParser.WatchContext)
        assert isinstance(watch, pcodeParser.WatchContext)
        mark = get_first_child(program, pcodeParser.MarkContext)
        self.assertIsInstance(mark, pcodeParser.MarkContext)

    def test_watch_w_inst_error_minimal(self):
        p = parse("""
Watch
    Mark: A""")
        program = p.parser.program()  # type: ignore
        # p.printSyntaxTree(program)
        watch = get_first_child(program, pcodeParser.WatchContext)
        assert isinstance(watch, pcodeParser.WatchContext)
        mark = get_first_child(program, pcodeParser.MarkContext)
        self.assertIsInstance(mark, pcodeParser.MarkContext)

    def test_condition_unit(self):
        p = parse("X > 10 mL")
        c = p.parser.condition()  # type: ignore
        self.assertIsNotNone(c)
        self.assertIsInstance(c, pcodeParser.ConditionContext)
        self.assertConditionValue(c, "X", ">", "10 mL")
        self.assertContextHasNoChildError(c)

    def test_condition_unit_no_ws(self):
        p = parse("X > 10mL")
        c = p.parser.condition()  # type: ignore
        self.assertIsNotNone(c)
        self.assertIsInstance(c, pcodeParser.ConditionContext)
        self.assertConditionValue(c, "X", ">", "10mL")
        self.assertContextHasNoChildError(c)

    def test_condition_unit_percentage(self):
        p = parse("X > 10 %")
        c = p.parser.condition()  # type: ignore
        self.assertIsNotNone(c)
        self.assertIsInstance(c, pcodeParser.ConditionContext)
        self.assertConditionValue(c, "X", ">", "10 %")
        self.assertContextHasNoChildError(c)

    def test_condition_ws(self):
        p = parse("X> 10mL")
        c = p.parser.condition()  # type: ignore
        self.assertIsNotNone(c)
        # p.printSyntaxTree(c)
        self.assertIsInstance(c, pcodeParser.ConditionContext)
        self.assertConditionValue(c, "X", ">", "10mL")
        self.assertContextHasNoChildError(c)

    def test_identifier(self):
        def test(tag: str):
            with self.subTest(tag):
                p = parse(tag)
                c = p.parser.identifier()  # type: ignore
                self.assertIsNotNone(c)
                # p.printSyntaxTree(c)
                self.assertIsInstance(c, pcodeParser.IdentifierContext)
                self.assertContextHasNoChildError(c)
        test("X2")
        test("XY2")
        test("x2")
        test("foo")
        test("foo bar")

        test("KaZ")
        test("e 2")
        test("e")
        test("X")
        test("xY")
        test("Xy")

        # these are also valid identifiers even though they are also units
        test("s")
        test("L")
        test("min")
        test("mL")

    def test_identifier_ext(self):
        def test(tag: str):
            with self.subTest(tag):
                p = parse(tag)
                c = p.parser.identifier_ext()  # type: ignore
                self.assertIsNotNone(c)
                # p.printSyntaxTree(c)
                self.assertIsInstance(c, pcodeParser.Identifier_extContext)
                self.assertContextHasNoChildError(c)
                self.assertEqual(tag, c.getText())
        test("X2")
        test("XY2")
        test("x2")
        test("foo")
        test("foo bar")

        test("KaZ")
        test("e 2")
        test("e")
        test("X")
        test("xY")
        test("Xy")

        # these are also valid identifiers even though they are also units
        test("s")
        test("L")
        test("min")
        test("mL")

        # stuff that is not an identifier
        test("87")
        test("_")
        test(" 87: _")
        test("s+")
        test("B %")
        test("5%: test")

    def test_condition_tag_ws(self):
        p = parse("Block Time > 0.2 min")
        c = p.parser.condition()  # type: ignore
        self.assertIsNotNone(c)
        # p.printSyntaxTree(c)
        self.assertIsInstance(c, pcodeParser.ConditionContext)
        self.assertConditionValue(c, "Block Time", ">", "0.2 min")
        self.assertContextHasNoChildError(c)

    def test_condition_unitless(self):
        p = parse("X > 10")
        c = p.parser.condition()  # type: ignore
        self.assertIsNotNone(c)
        # p.printSyntaxTree(c)
        self.assertIsInstance(c, pcodeParser.ConditionContext)
        self.assertConditionValue(c, "X", ">", "10")
        self.assertContextHasNoChildError(c)

    def test_condition_negative_value(self):
        p = parse("X > -10")
        c = p.parser.condition()  # type: ignore
        self.assertIsNotNone(c)
        # p.printSyntaxTree(c)
        self.assertIsInstance(c, pcodeParser.ConditionContext)
        self.assertConditionValue(c, "X", ">", "-10")
        self.assertContextHasNoChildError(c)

    def test_condition_error(self):
        p = parse("X > ab mL")
        c = p.parser.condition()  # type: ignore
        self.assertIsNotNone(c)
        # p.printSyntaxTree(c)
        self.assertIsInstance(c, pcodeParser.ConditionContext)
        # this is not a parse error - will fail during build
        self.assertConditionValue(c, "X", ">", "ab mL")

    def test_condition_error_2(self):
        p = parse("X>-ab mL")
        c = p.parser.condition()  # type: ignore
        self.assertIsNotNone(c)
        # p.printSyntaxTree(c)
        self.assertIsInstance(c, pcodeParser.ConditionContext)
        # this is not a parse error - will fail during build
        self.assertConditionValue(c, "X", ">", "-ab mL")

    def assertIsWatchWithCondition(self, c: ParserRuleContext, condition: str):
        self.assertIsNotNone(c)
        self.assertIsInstance(c, pcodeParser.WatchContext)
        _args = None
        if c.children is not None:
            for x in c.children:
                if isinstance(x, pcodeParser.ConditionContext):
                    _args = x.getText()
        self.assertEqual(_args, condition)

    def assertConditionValue(self, c: pcodeParser.ConditionContext,
                             expected_lhs, expected_op, expected_rhs):
        lhs, op, rhs = None, None, None
        if c.children is not None:
            for x in c.children:
                if isinstance(x, pcodeParser.Condition_lhsContext):
                    lhs = x.getText()
                elif isinstance(x, pcodeParser.Compare_opContext):
                    op = x.getText()
                elif isinstance(x, pcodeParser.Condition_rhsContext):
                    rhs = x.getText()
        self.assertEqual(lhs, expected_lhs)
        self.assertEqual(op, expected_op)
        self.assertEqual(rhs, expected_rhs)

    def test_watch_with_condition(self):
        p = parse("Watch: X > 10 mL")
        c = p.parser.watch()  # type: ignore
        self.assertIsNotNone(c)
        # p.printSyntaxTree(c)
        self.assertIsWatchWithCondition(c, "X > 10 mL")
        self.assertConditionValue(c.condition(), "X", ">", "10 mL")
        self.assertContextHasNoChildError(c)

    def assertIsAlarmWithCondition(self, c: ParserRuleContext, condition: str):
        self.assertIsNotNone(c)
        self.assertIsInstance(c, pcodeParser.AlarmContext)
        _args = None
        if c.children is not None:
            for x in c.children:
                if isinstance(x, pcodeParser.ConditionContext):
                    _args = x.getText()
        self.assertEqual(_args, condition)

    def test_alarm(self):
        p = parse("Alarm")
        c = p.parser.alarm()  # type: ignore
        self.assertIsNotNone(c)
        # p.printSyntaxTree(c)
        self.assertIsInstance(c, pcodeParser.AlarmContext)
        self.assertContextHasNoChildError(c)

    def test_alarm_with_condition(self):
        p = parse("Alarm: X > 10 mL")
        c = p.parser.alarm()  # type: ignore
        condition = c.condition()
        assert isinstance(condition, pcodeParser.ConditionContext)
        self.assertIsNotNone(c)
        # p.printSyntaxTree(c)
        self.assertIsAlarmWithCondition(c, "X > 10 mL")
        self.assertConditionValue(condition, "X", ">", "10 mL")
        self.assertContextHasNoChildError(c)

    def test_increment_rc(self):
        p = parse("Increment run counter")
        c = p.parser.increment_rc()  # type: ignore
        self.assertIsNotNone(c)
        # p.printSyntaxTree(c)
        self.assertIsInstance(c, pcodeParser.Increment_rcContext)
        self.assertContextHasNoChildError(c)

    def test_stop(self):
        p = parse("Stop")
        c = p.parser.stop()  # type: ignore
        self.assertIsNotNone(c)
        # p.printSyntaxTree(c)
        self.assertIsInstance(c, pcodeParser.StopContext)
        self.assertContextHasNoChildError(c)

    def test_pause(self):
        p = parse("Pause")
        c = p.parser.pause()  # type: ignore
        self.assertIsNotNone(c)
        # p.printSyntaxTree(c)
        self.assertIsInstance(c, pcodeParser.PauseContext)
        self.assertContextHasNoChildError(c)

    def test_pause_w_arg(self):
        p = parse("Pause: 1 min")
        c = p.parser.pause()  # type: ignore
        self.assertIsNotNone(c)
        # p.printSyntaxTree(c)
        self.assertIsInstance(c, pcodeParser.PauseContext)
        self.assertContextHasNoChildError(c)

    def test_hold(self):
        p = parse("Hold")
        c = p.parser.hold()  # type: ignore
        self.assertIsNotNone(c)
        # p.printSyntaxTree(c)
        self.assertIsInstance(c, pcodeParser.HoldContext)
        self.assertContextHasNoChildError(c)

    def test_hold_w_arg(self):
        p = parse("Hold: 1 min")
        c = p.parser.hold()  # type: ignore
        self.assertIsNotNone(c)
        # p.printSyntaxTree(c)
        self.assertIsInstance(c, pcodeParser.HoldContext)
        self.assertContextHasNoChildError(c)

    def test_wait_w_arg(self):
        p = parse("Wait:1s")
        c = p.parser.wait()  # type: ignore
        self.assertIsNotNone(c)
        # p.printSyntaxTree(c)
        self.assertIsInstance(c, pcodeParser.WaitContext)
        self.assertContextHasNoChildError(c)

    def test_wait(self):
        # Note: Wait without duration is not a valid command, but its validity
        # is checked post-parsing so parsing itself must succeed
        p = parse("Wait")
        c = p.parser.wait()  # type: ignore
        self.assertIsNotNone(c)
        # p.printSyntaxTree(c)
        self.assertIsInstance(c, pcodeParser.WaitContext)
        self.assertContextHasNoChildError(c)

    def test_duration(self):
        def test(code):
            with self.subTest(code):
                p = parse(code)
                c = p.parser.duration()  # type: ignore
                self.assertIsNotNone(c)
                # p.printSyntaxTree(c)
                self.assertIsInstance(c, pcodeParser.DurationContext)
                self.assertContextHasNoChildError(c)

        test("1 min")
        test("1.0 min")
        test("131,654 min")
        test("1.0 s")
        test("1s")
        test("87 h")
        test("87 foo")


    def test_restart(self):
        p = parse("Restart")
        c = p.parser.restart()  # type: ignore
        self.assertIsNotNone(c)
        # p.printSyntaxTree(c)
        self.assertIsInstance(c, pcodeParser.RestartContext)
        self.assertContextHasNoChildError(c)

    def assertIsPCommandWithNameArgs(self, c: ParserRuleContext, name: str, args: str):
        self.assertIsNotNone(c)
        self.assertIsInstance(c, pcodeParser.CommandContext)

        _name = None
        _args = None

        if c.children is not None:
            for x in c.children:
                if isinstance(x, pcodeParser.Command_nameContext):
                    _name = x.getText()
                elif isinstance(x, pcodeParser.Command_argsContext):
                    _args = x.getText()

        self.assertEqual(_name, name)
        self.assertEqual(_args, args)

    def test_command_name(self):
        self.assertEqual(parse("foo").parser.command_name().getText(), "foo")  # type: ignore

    def test_command_args(self):
        def test(code, expected_args):
            with self.subTest(code):
                # print(f"\nCode: '{code}'")
                p = parse(code)
                c = p.parser.command_args()  # type: ignore
                # print("Tree:")
                # p.printSyntaxTree(c)
                self.assertEqual(c.getText(), expected_args)

        test("bar", "bar")
        test("bar baz", "bar baz")
        test("bar=baz", "bar=baz")
        test("bar,baz", "bar,baz")
        # TODO test("bar,baz #comment", "bar,baz")

    def test_command(self):
        def test(code):
            with self.subTest(code):
                c = parse(code).parser.command()  # type: ignore
                self.assertIsNotNone(c)
                self.assertIsInstance(c, pcodeParser.CommandContext)
                # self.assertContextHasNoChildError(c)

        test("foo")
        test(" foo")
        test("   foo")
        test("1 foo")
        test("1.1 foo")
        test("1.02 foo")
        test("  1.3 foo")
        test("   1.02 foo")

        test("foo:")
        test(" foo:a:3")
        test(" foo: a:3")
        test("   foo:a=3")
        test("1 foo:3,4 ")
        test("1.1 foo: 3=4,5:6")
        test("2.95 Zero UV")

    def test_command_name_args_simple(self):
        def test(code, expected_name, expected_args):
            with self.subTest(code):
                # print(f"\nCode: '{code}'")
                p = parse(code)
                c = p.parser.command()  # type: ignore
                # print("Tree:")
                # p.printSyntaxTree(c)
                self.assertIsPCommandWithNameArgs(c, expected_name, expected_args)
                self.assertContextHasNoChildError(c)

        test("foo", "foo", None)
        test("foo: bar", "foo", "bar")
        test("1 foo: bar", "foo", "bar")
        test("7.89 foo: bar", "foo", "bar")
        test("7,89 foo: bar", "foo", "bar")
        test("2.95 Zero UV", "Zero UV", None)

    def test_command_name_args(self):
        def test(code, expected_name, expected_args):
            with self.subTest(code):
                # print(f"\nCode: '{code}'")
                p = parse(code)
                c = p.parser.command()  # type: ignore
                # print("Tree:")
                # p.printSyntaxTree(c)
                self.assertIsPCommandWithNameArgs(c, expected_name, expected_args)
                self.assertContextHasNoChildError(c)

        test("1.23 foo bar", "foo bar", None)
        test("foo: bar,baz", "foo", "bar,baz")
        test("1.23 foo: bar baz", "foo", "bar baz")
        test("foo: 3,5", "foo", "3,5")
        test("foo: 3=5", "foo", "3=5")

        test("Inlet PU01: VA01", "Inlet PU01", "VA01")
        test("PU01: 77.64 %", "PU01", "77.64 %")

    def test_program_ws(self):
        code = """
Block
    Mark : 1
        """
        p = parse(code)
        ctx = p.parser.program()  # type: ignore

        # p.printSyntaxTree(ctx)
        self.assertContextHasNoChildError(ctx)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import asyncio
import logging
import random
import time
from typing import Awaitable, Callable, Literal

from openpectus.engine.engine_message_builder import EngineMessageBuilder
from openpectus.lang.exec.events import EventEmitter, EventListener
from openpectus.protocol.engine_dispatcher import EngineDispatcher
import openpectus.protocol.engine_messages as EM
from openpectus.protocol.exceptions import ProtocolNetworkException
import openpectus.protocol.messages as M

logger = logging.getLogger(__name__)

MAX_RECONNECT_WAIT_SECONDS = 10
BUFFER_BATCH_SIZE = 10

RecoverState = Literal[
    "Started",
    "Connected",
    "Failed",
    "Disconnected", "Reconnecting", "CatchingUp", "Reconnected",
    "Stopped",
    "ShutdownComplete"
]

AsyncConnectionCallback = Callable[[], Awaitable[None]]
StateChangingCallback = Callable[[RecoverState, RecoverState], Awaitable[None]]

class AsyncTimer:
    def __init__(self, timeout, callback):
        self._timeout = timeout
        self._callback = callback
        self._task = asyncio.create_task(self._job(), name="engine.engine_runner.AsyncTimer")

    async def _job(self):
        try:
            while True:
                await asyncio.sleep(self._timeout)
                await self._callback()
        except asyncio.CancelledError:
            logger.debug("Cancelled AsyncTimer")

    def start(self):
        pass

    def stop(self):
        self._task.cancel()

class EngineRunner(EventListener):
    """
    Runner that implements error recovery and masks connection errors that may occur in the Aggregator-Engine Protocol.
    """

    # TODO: Remove raising ProtocolException where they are not caught which is most of them

    def __init__(self,
                 dispatcher: EngineDispatcher,
                 message_builder: EngineMessageBuilder,
                 emitter: EventEmitter,
                 loop: asyncio.AbstractEventLoop) -> None:
        super().__init__()
        self._dispatcher = dispatcher
        self._message_builder = message_builder
        self._state: RecoverState = "Started"
        self._message_buffer: list[EM.EngineMessage] = []
        self._lock = asyncio.Lock()
        self._state_task: asyncio.Task[None] | None = None
        self._timer = AsyncTimer(0.1, self._tick)
        self._first_steady_state = True
        self._loop = loop

        emitter.add_listener(self)

        self.connected_callback: AsyncConnectionCallback | None = None
        self.failed_callback: AsyncConnectionCallback | None = None
        self.catchingup_callback: AsyncConnectionCallback | None = None
        self.reconnecting_callback: AsyncConnectionCallback | None = None
        self.reconnected_callback: AsyncConnectionCallback | None = None
        self.state_changing_callback: StateChangingCallback | None = None
        self.first_steady_state_callback: AsyncConnectionCallback | None = None

    def on_start(self, run_id: str):
        super().on_start(run_id)
        if True or self.state == "Connected" or self.state == "Reconnected":
            msg = self._message_builder.create_run_started_msg(run_id, time.time())
            try:
                self.post(msg)
            except Exception:
                logger.warning("Failed to send RunStartedMsg, adding to buffer")
                self._buffer_message(msg)

    def on_stop(self):
        assert self.run_id is not None
        run_id = self.run_id
        super().on_stop()
        msg = self._message_builder.create_run_stopped_msg(run_id)
        try:
            self.post(msg)
        except Exception:
            logger.warning("Failed to send RunStoppedMsg, adding to buffer")
            self._buffer_message(msg)

    @property
    def state(self) -> RecoverState:
        return self._state

    async def run(self):
        self._timer.start()

        while self.state != "ShutdownComplete":
            await asyncio.sleep(0.2)
        self._timer.stop()
        await self._timer._task

    async def _connect_async(self):
        if self.state == "Stopped":
            return
        elif self.state == "Started" or self.state == "Disconnected":
            pass
        else:
            err_msg = f"Invalid operation '_connect_async' for state: {self.state}"
            logger.error(err_msg)

        try:
            await self._dispatcher.connect_async()

            # Note: Consider adding a ping to verify connection.
            # resp = await self._dispatcher.post_async(EM.PingMsg())
            # if resp is None or not isinstance(resp, M.SuccessMessage):
            #     raise Exception("Negative Ping result")

            if self.state == "Started":
                await self._set_state("Connected")
            elif self.state == "Disconnected":
                await self._set_state("Reconnecting")
            logger.info("Connect successful")

        except ProtocolNetworkException:
            await self._set_state("Failed")
            logger.warning("Connect failed")
        except Exception:
            await self._set_state("Failed")
            logger.error("Connect failed with unhandled exception", exc_info=True)

    async def _disconnect_async(self, set_state_disconnected=True):
        if self.state == "Stopped":
            pass
        elif self.state == "Failed":
            pass
        else:
            err_msg = f"Invalid operation '_disconnect_async' for state: {self.state}"
            logger.error(err_msg)

        try:
            await self._dispatcher.disconnect_async()
        except Exception:
            logger.warning("Disconnect failed")
        finally:
            if set_state_disconnected:
                await self._set_state("Disconnected")

    async def shutdown(self):
        await self._set_state("Stopped")
        await self._disconnect_async(set_state_disconnected=False)
        await self._set_state("ShutdownComplete")

    def post(self, message: EM.EngineMessage) -> M.MessageBase:
        # send sync message using self._loop
        states_to_post: list[RecoverState] = ["Connected", "Reconnected"]
        assert self.state in states_to_post, f"Invalid Post state for message: {message.ident}"

        task = asyncio.run_coroutine_threadsafe(self._dispatcher.send_async(message), self._loop)
        while not task.done():
            time.sleep(0.01)
        ex = task.exception()
        if ex is None:
            return M.SuccessMessage()
        else:
            logger.error(f"Failed to send message: {ex}")
            return M.ErrorMessage(message="Failed to send message")

    async def _post_async(self, message: EM.EngineMessage) -> M.MessageBase:
        states_to_post: list[RecoverState] = ["Connected", "Reconnected", "CatchingUp"]
        states_to_buffer: list[RecoverState] = ["Failed", "Disconnected", "Reconnecting"]

        if self.state == "Stopped":
            return M.ErrorMessage(message="System is down")

        elif self.state in states_to_post:
            try:
                return await self._dispatcher.send_async(message)
            except ProtocolNetworkException:
                await self._set_state("Failed")
                logger.warning(f"Failed to send message {message.ident}, adding to buffer")
                self._buffer_message(message)
                return M.ErrorMessage(message="Failed to send message, was added to buffer")

        elif self.state in states_to_buffer:
            self._buffer_message(message)
            return M.ErrorMessage(message="Failed to send message, was added to buffer")

        else:
            err_msg = f"Invalid operation 'post_async' for state: {self.state}"
            logger.error(err_msg)
            return M.ErrorMessage(message="Failed to send message, invalid state")

    async def _tick(self):
        async with self._lock:
            if self._state == "Stopped":
                return

            elif self._state == "Connected" or self._state == "Reconnected":
                # steady state
                pass

            elif self._state == "Started":
                await self._connect_async()

            elif self._state == "Failed":
                await self._disconnect_async()

            elif self._state == "Disconnected":
                wait_secs = random.uniform(0.5, MAX_RECONNECT_WAIT_SECONDS)
                logger.debug(f"Waiting {wait_secs:0.1f} seconds before reconnecting")
                await asyncio.sleep(wait_secs)
                await self._connect_async()

            elif self._state == "Reconnecting":
                await self._set_state("CatchingUp")

            elif self._state == "CatchingUp":
                await self._send_buffered_batch()

    async def _set_state(self, state: RecoverState):
        prev_state = self.state
        logger.info(f"Changing state: {prev_state} -> {state}")
        if self.state_changing_callback is not None:
            await self.state_changing_callback(self.state, state)

        self._state = state

        if self._state_task is not None:
            # cancel the long running task, except if its the fail/buffer task and we're still
            # working to reconnect
            if (state == "Failed" or state == "Disconnected" or state == "Reconnecting"
                    or state == "CatchingUp") and self._state_task.get_name() == "buffer_messages":
                pass
            else:
                logger.debug("Cancelling state task " + self._state_task.get_name())
                self._state_task.cancel()
                try:
                    await self._state_task  # Await cancellation
                except asyncio.CancelledError:
                    pass
                self._state_task = None

        if state == "Connected":
            await self._on_connected()
        elif state == "Failed":
            self._dispatcher._engine_id = None
            await self._on_failed()
        elif state == "CatchingUp":
            await self._on_catchingup()
        elif state == "Reconnecting":
            await self._on_reconnecting()
        elif state == "Reconnected":
            await self._on_reconnected()

    async def _send_buffered_batch(self):
        if self._state == "Stopped":
            return

        if self._state != "CatchingUp":
            err_msg = f"Invalid operation '_send_buffered_batch' for state: {self.state}"
            logger.warning(err_msg)

        buffered_message_count = self._get_buffer_size()
        if buffered_message_count > 0:
            message_batch = self._pop_buffer_batch()
            if __debug__:
                logger.debug(f"Buffered messages remaining to be sent: {buffered_message_count}")
                logger.debug(f"Buffer: {EM.print_sequence_range(self._message_buffer)}")
                logger.debug(f"Batch : {EM.print_sequence_range(message_batch)}")

            for message in message_batch:
                await self._post_async(message)
            logger.debug("Done sending batch")
        else:
            logger.info("All caught up sending buffered messages")
            await self._set_state("Reconnected")

    def _buffer_message(self, message: EM.EngineMessage):
        """ Append message to the buffer """
        self._dispatcher.assign_sequence_number(message)
        logger.debug(f"Buffering message: {message.ident}")
        self._message_buffer.append(message)

    def _get_buffer_size(self) -> int:
        return len(self._message_buffer)

    def _pop_buffer_batch(self) -> list[EM.EngineMessage]:
        messages = []
        for _ in range(BUFFER_BATCH_SIZE):
            if len(self._message_buffer) > 0:
                messages.append(self._message_buffer.pop(0))
        return messages

    async def _on_connected(self):
        logger.debug("on_connected")
        if self.connected_callback is not None:
            await self.connected_callback()
        await self.on_steady_state()

    async def _on_failed(self):
        logger.debug("on_failed")

        async def buffer_messages():
            logger.info("Started buffer_messages loop")
            try:
                while True:
                    if self.state == "Stopped":
                        return
                    for msg in [
                        self._message_builder.create_tag_updates_msg(),
                        self._message_builder.create_method_state_msg(),
                        self._message_builder.create_error_log_msg(),
                    ]:
                        if msg is not None:
                            self._buffer_message(msg)
                    await asyncio.sleep(5)
            except asyncio.CancelledError:
                logger.debug("Cancelled _state_task")

        # need to run long running job in spawn task
        if self._state_task is None:
            self._state_task = asyncio.create_task(buffer_messages(), name="buffer_messages")

    async def _on_reconnecting(self):
        logger.debug("on_reconnecting")
        if self.reconnecting_callback is not None:
            await self.reconnecting_callback()

    async def _on_catchingup(self):
        logger.debug("on_catchingup")
        if self.catchingup_callback is not None:
            await self.catchingup_callback()

    async def _on_reconnected(self):
        logger.debug("on_reconnected")
        if self.reconnected_callback is not None:
            await self.reconnected_callback()
        await self.on_steady_state()

    async def on_steady_state(self):
        """ Steady-State message sending loop """

        # raise first steady state event that should start engine
        if self._first_steady_state:
            self._first_steady_state = False
            if self.first_steady_state_callback is not None:
                await self.first_steady_state_callback()

        async def send_messages():
            logger.info("Started steady-state sending loop")
            try:
                await self._post_async(self._message_builder.create_uod_info())
                await self._post_async(self._message_builder.create_tag_updates_snapshot_msg())
                while True:
                    messages = []
                    try:
                        messages = [
                            self._message_builder.create_control_state_msg(),
                            self._message_builder.create_method_state_msg(),
                            self._message_builder.create_error_log_msg(),
                            self._message_builder.create_tag_updates_msg(),
                        ]
                        if self.run_id is not None:
                            messages.append(self._message_builder.create_runlog_msg(self.run_id))
                    except Exception:
                        logger.error("Exception occurred building messages", exc_info=True)

                    for msg in messages:
                        if msg is not None:
                            await self._post_async(msg)
                    await asyncio.sleep(0.3)
            except asyncio.CancelledError:
                logger.info("Cancelled _state_task")
            except Exception:
                logger.error("Exception interrupted _state_task")

        # Send message in concurrent task
        self._state_task = asyncio.create_task(send_messages(), name="engine.engine_runner.on_steady_state.send_messages")

    def __str__(self) -> str:
        return f"EngineRunner(state: {self.state})"
